/*
 * Copyright (c) 2019,2023-2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                             Martin Jacquet - September 2022
 */
#ifndef H_RS_CODELS
#define H_RS_CODELS

#include "acrealsense.h"

#include <err.h>

#include <unordered_map>

#include "librealsense2/rs.hpp"
#include "camera/camera.h"
#include "realsense_c_types.h"

struct realsense_pipeline : rs2::pipeline {};
struct realsense_config : rs2::config {};
struct realsense_frame_queue : rs2::frame_queue {};
struct realsense_frame : rs2::frame {};

/* map rs2_stream and index to a job */
struct realsense_stream_id {
  realsense_stream_id(const rs2::stream_profile &p)
    : type(p.stream_type()), index(p.stream_index()) {}

  struct hash {
    std::size_t operator() (const realsense_stream_id &i) const {
      return i.index | i.type << 8;
    }
  };

  bool operator==(const realsense_stream_id &other) const {
    return type == other.type && index == other.index;
  }

  rs2_stream type;
  int index;
};

struct realsense_job : camera_job {
  void destroy() { reset(); }
  bool runnable() const { return frames.size() > 0; }
  state operator()();

  rs2::frame_queue frames{1};
  rs2::stream_profile p0;
  or_sensor_pixmap_encoding format;
};


struct realsense_streammap : std::unordered_map<
  realsense_stream_id, realsense_job, realsense_stream_id::hash> {};


/* map rs2_stream and realsense_stream enums */
static inline realsense_stream
rscvt_stream(const rs2_stream &s)
{
  switch(s) {
    case RS2_STREAM_COLOR: return realsense_STREAM_COLOR;
    case RS2_STREAM_INFRARED: return realsense_STREAM_INFRARED;
    case RS2_STREAM_DEPTH: return realsense_STREAM_DEPTH;
    case RS2_STREAM_GYRO: return realsense_STREAM_GYRO;
    case RS2_STREAM_ACCEL: return realsense_STREAM_ACCEL;

    default: return realsense_STREAM_INVALID;
  }
}

static inline rs2_stream
rscvt_stream(const realsense_stream &s)
{
  switch(s) {
    case realsense_STREAM_COLOR: return RS2_STREAM_COLOR;
    case realsense_STREAM_INFRARED: return RS2_STREAM_INFRARED;
    case realsense_STREAM_DEPTH: return RS2_STREAM_DEPTH;
    case realsense_STREAM_GYRO: return RS2_STREAM_GYRO;
    case realsense_STREAM_ACCEL: return RS2_STREAM_ACCEL;

    default: return RS2_STREAM_COUNT;
  }
}


/* map rs2_format and or_sensor_pixmap_encoding enums */
static inline or_sensor_pixmap_encoding
rscvt_format(const rs2_format &f)
{
  switch(f) {
    case RS2_FORMAT_ANY: return or_sensor_pixmap_ANY;
    case RS2_FORMAT_Y8: return or_sensor_pixmap_MONO8;
    case RS2_FORMAT_Y16: return or_sensor_pixmap_MONO16;
    case RS2_FORMAT_RGB8: return or_sensor_pixmap_RGB8;
    case RS2_FORMAT_BGR8: return or_sensor_pixmap_BGR8;
    case RS2_FORMAT_RGBA8: return or_sensor_pixmap_RGBA8;
    case RS2_FORMAT_BGRA8: return or_sensor_pixmap_BGRA8;

    default: return or_sensor_pixmap_INVALID;
  }
}

static inline rs2_format
rscvt_format(const or_sensor_pixmap_encoding &f)
{
  switch(f) {
    case or_sensor_pixmap_ANY: return RS2_FORMAT_ANY;
    case or_sensor_pixmap_MONO8: return RS2_FORMAT_Y8;
    case or_sensor_pixmap_MONO16: return RS2_FORMAT_Y16;
    case or_sensor_pixmap_BGR8: return RS2_FORMAT_BGR8;

    default: return RS2_FORMAT_COUNT;
  }
}


/* --- sequence operator|= ------------------------------------------------- */

/* Add unique element to bounded sequence with |= */
template<class T, typename E> void
operator|=(T &s, E &&x)
{
  uint32_t i;

  for(i = 0; i < s._length; i++)
    if (s._buffer[i] == x) return;

  if (s._length >= s._maximum) {
    std::ostringstream m;
    m << "too many elements, output '" << x << "' not included";
    warnx("%s", m.str().c_str());
    return;
  }

  s._buffer[s._length++] = x;
}

static inline bool
operator==(const realsense_video_size &rhs, const realsense_video_size &lhs)
{
  return rhs.width == lhs.width && rhs.height == lhs.height;
}

static inline std::ostream &
operator<<(std::ostream &s, const realsense_video_size &sz)
{
  return s << sz.width << ", " << sz.height;
}

#endif /* H_RS_CODELS */
