/*
 * Copyright (c) 2019,2023-2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                             Martin Jacquet - September 2022
 */

#include "acrealsense.h"

#include <err.h>

#include <sstream>

#include "realsense_c_types.h"

#include "codels.hh"


static genom_event	rs_stream_enable(realsense_stream type, int16_t index,
                                int16_t width, int16_t height,
                                or_sensor_pixmap_encoding format, int16_t fps,
                                realsense_config **config,
                                realsense_pipeline **rspipe,
                                const char name[16],
                                const camera_pipeline *pipeline,
                                const realsense_frame_queue *mqueue,
                                realsense_streammap **streammap,
                                const genom_context self);
static genom_event	rs_stream_start(realsense_config **config,
                                realsense_pipeline **rspipe,
                                const camera_pipeline *pipeline,
                                const realsense_frame_queue *mqueue,
                                realsense_streammap **streammap,
                                const genom_context self);


/* --- Function get_video_stream_formats -------------------------------- */

/** Validation codel rs_is_video of function get_video_stream_formats.
 *
 * Returns genom_ok.
 * Throws realsense_einval.
 */
genom_event
rs_is_video(realsense_stream type, const genom_context self)
{
  switch (type) {
    default: break;

    case realsense_STREAM_ANY:
    case realsense_STREAM_DEPTH:
    case realsense_STREAM_GYRO:
    case realsense_STREAM_ACCEL:
    case realsense_STREAM_INVALID:
      /* invalid type */
      return realsense_einval(self);
  }

  return genom_ok;
}


/* --- Function get_depth_stream_formats -------------------------------- */

/** Validation codel rs_is_depth of function get_depth_stream_formats.
 *
 * Returns genom_ok.
 * Throws realsense_einval.
 */
genom_event
rs_is_depth(realsense_stream type, const genom_context self)
{
  switch (type) {
    default: break;

    case realsense_STREAM_ANY:
    case realsense_STREAM_COLOR:
    case realsense_STREAM_INFRARED:
    case realsense_STREAM_GYRO:
    case realsense_STREAM_ACCEL:
    case realsense_STREAM_INVALID:
      /* invalid type */
      return realsense_einval(self);
  }

  return genom_ok;
}


/* --- Function get_video_stream_sizes ---------------------------------- */

/** Validation codel rs_is_video of function get_video_stream_sizes.
 *
 * Returns genom_ok.
 * Throws realsense_einval.
 */
/* already defined in service get_video_stream_formats validation */



/* --- Function get_depth_stream_sizes ---------------------------------- */

/** Validation codel rs_is_depth of function get_depth_stream_sizes.
 *
 * Returns genom_ok.
 * Throws realsense_einval.
 */
/* already defined in service get_depth_stream_formats validation */



/* --- Function stream_video -------------------------------------------- */

/** Validation codel rs_bank_validate of function stream_video.
 *
 * Returns genom_ok.
 * Throws realsense_einval, realsense_e_rs.
 */
genom_event
rs_bank_validate(const char name[16], const genom_context self)
{
  if (!*name) return realsense_einval(self);
  // if (slot >= or_image_max_slots) return realsense_einval(self);

  return genom_ok;
}

/** Validation codel rs_is_video of function stream_video.
 *
 * Returns genom_ok.
 * Throws realsense_einval, realsense_e_rs.
 */
/* already defined in service get_video_stream_formats validation */



/* --- Function stream_depth -------------------------------------------- */

/** Validation codel rs_bank_validate of function stream_depth.
 *
 * Returns genom_ok.
 * Throws realsense_einval, realsense_e_rs.
 */
/* already defined in service stream_video validation */


/** Validation codel rs_is_depth of function stream_depth.
 *
 * Returns genom_ok.
 * Throws realsense_einval, realsense_e_rs.
 */
/* already defined in service get_depth_stream_formats validation */



/* --- Function stream_motion ------------------------------------------- */

/** Validation codel rs_is_motion of function stream_motion.
 *
 * Returns genom_ok.
 * Throws realsense_einval, realsense_e_rs.
 */
genom_event
rs_is_motion(realsense_stream type, const genom_context self)
{
  switch (type) {
    default: break;

    case realsense_STREAM_ANY:
    case realsense_STREAM_COLOR:
    case realsense_STREAM_INFRARED:
    case realsense_STREAM_DEPTH:
    case realsense_STREAM_INVALID:
      /* invalid type */
      return realsense_einval(self);
  }

  return genom_ok;
}


/* --- Function cancel -------------------------------------------------- */

/** Codel rs_cancel of function cancel.
 *
 * Returns genom_ok.
 * Throws camera_einval, realsense_e_rs.
 */
genom_event
rs_cancel(realsense_config **config, realsense_pipeline **rspipe,
          const camera_pipeline *pipeline,
          const camera_worker_lock *mutex,
          const realsense_frame_queue *mqueue,
          realsense_streammap **streammap, const genom_context self)
{
  /* check if one job got affected */
  bool changed = false;
  for(const auto &kv : **streammap)
    if (!kv.second.nout()) changed = true;
  if (!changed) return genom_ok;

  rs2::pipeline_profile &&p = (*rspipe)->get_active_profile();

  /* There is an issue with config->disable_stream() that prevents a stream to
   * be enabled again when using default index (-1). So here all streams are
   * disabled and those still active are re-added */
  bool empty = true;
  (*config)->disable_all_streams();

  /* motion streams */
  for(const auto &s : p.get_streams())
    switch(s.stream_type()) {
      case RS2_STREAM_GYRO:
      case RS2_STREAM_ACCEL:
        (*config)->enable_stream(s.stream_type(), s.stream_index());
        empty = false;
        break;
      default: break;
    }

  /* image streams */
  for(const auto &kv : **streammap) {
    const realsense_stream_id &i = kv.first;
    const realsense_job &j = kv.second;

    if (!j.nout()) { /* disable this stream */
      if (p) warnx("disabled %s stream",
                   p.get_stream(i.type, i.index).stream_name().c_str());
      continue;
    }

    (*config)->enable_stream(i.type, i.index);
    empty = false;
  }

  /* stop to re-apply the new config */
  try { (*rspipe)->stop(); }
  catch(const rs2::wrong_api_call_sequence_error &) {
    /* OK: not started yet, nothing to do */
    return genom_ok;
  }

  /* restart devices */
  return !empty ?
    rs_stream_start(config, rspipe, pipeline, mqueue, streammap, self) :
    genom_ok;
}


/* --- Function get_streams --------------------------------------------- */

/** Codel rs_get_streams of function get_streams.
 *
 * Returns genom_ok.
 */
genom_event
rs_get_streams(sequence16_realsense_stream *streams,
               const genom_context self)
{
  rs2::context ctx;
  realsense_stream t;

  streams->_length = 0;
  try {
    for(const auto &s : ctx.query_all_sensors())
      for(const auto &p : s.get_stream_profiles()) {
        /* filter unknown types */
        t = rscvt_stream(p.stream_type());
        if (t == realsense_STREAM_INVALID) continue;

        *streams |= t;
      }
  } catch(const rs2::error &e) { warnx("%s", e.what()); }

  return genom_ok;
}


/* --- Function get_stream_indices -------------------------------------- */

/** Codel rs_get_stream_indices of function get_stream_indices.
 *
 * Returns genom_ok.
 */
genom_event
rs_get_stream_indices(realsense_stream type, sequence16_long *indices,
                      const genom_context self)
{
  rs2::context ctx;

  if (rscvt_stream(type) == RS2_STREAM_COUNT) {
    /* invalid type */
    return realsense_einval(self);
  }

  indices->_length = 0;
  for(const auto &s : ctx.query_all_sensors())
    for(const auto &p : s.get_stream_profiles()) {
      /* match type */
      if (p.stream_type() != rscvt_stream(type)) continue;

      *indices |= p.stream_index();
    }

  return genom_ok;
}


/* --- Function get_stream_fps ------------------------------------------ */

/** Codel rs_get_stream_fps of function get_stream_fps.
 *
 * Returns genom_ok.
 */
genom_event
rs_get_stream_fps(realsense_stream type, sequence16_long *fps,
                  const genom_context self)
{
  rs2::context ctx;

  if (rscvt_stream(type) == RS2_STREAM_COUNT) {
    /* invalid type */
    return realsense_einval(self);
  }

  fps->_length = 0;
  for(const auto &s : ctx.query_all_sensors())
    for(const auto &p : s.get_stream_profiles()) {
      /* match type */
      if (p.stream_type() != rscvt_stream(type)) continue;

      *fps |= p.fps();
    }

  return genom_ok;
}


/* --- Function get_video_stream_formats -------------------------------- */

/** Codel rs_get_video_stream_formats of function get_video_stream_formats.
 *
 * Returns genom_ok.
 * Throws realsense_einval.
 */
genom_event
rs_get_video_stream_formats(realsense_stream type,
                            sequence16_or_sensor_pixmap_encoding *format,
                            const genom_context self)
{
  rs2::context ctx;
  or_sensor_pixmap_encoding f;

  format->_length = 0;
  for(const auto &s : ctx.query_all_sensors())
    for(const auto &p : s.get_stream_profiles()) {
      /* match type */
      if (p.stream_type() != rscvt_stream(type)) continue;

      /* filter unknown format */
      f = rscvt_format(p.format());
      if (f == or_sensor_pixmap_INVALID) continue;

      *format |= f;
    }

  return genom_ok;
}


/* --- Function get_depth_stream_formats -------------------------------- */

/** Codel rs_get_depth_stream_formats of function get_depth_stream_formats.
 *
 * Returns genom_ok.
 * Throws realsense_einval.
 */
genom_event
rs_get_depth_stream_formats(realsense_stream type,
                            sequence16_or_sensor_pixmap_encoding *format,
                            const genom_context self)
{
  rs2::context ctx;

  /* handle regular video streams */
  if (!rs_is_video(type, self))
    return rs_get_video_stream_formats(type, format, self);

  /* expose our own formats */
  format->_length = 0;
  for(const auto &s : ctx.query_all_sensors())
    for(const auto &p : s.get_stream_profiles()) {
      /* match type */
      if (p.stream_type() != RS2_STREAM_DEPTH) continue;

      *format |= or_sensor_pixmap_Z32;
      *format |= or_sensor_pixmap_XYZ32;
    }

  return genom_ok;
}


/* --- Function get_video_stream_sizes ---------------------------------- */

/** Codel rs_get_stream_sizes of function get_video_stream_sizes.
 *
 * Returns genom_ok.
 * Throws realsense_einval.
 */
genom_event
rs_get_stream_sizes(realsense_stream type,
                    sequence16_realsense_video_size *sizes,
                    const genom_context self)
{
  rs2::context ctx;

  sizes->_length = 0;
  for(const auto &s : ctx.query_all_sensors())
    for(const auto &p : s.get_stream_profiles()) {
      /* match type */
      if (!rs_is_depth(type, self)
          && p.stream_type() != rscvt_stream(type))
        continue;

      const auto &vp = p.as<rs2::video_stream_profile>();
      if (!vp) continue;

      *sizes |= realsense_video_size{vp.width(), vp.height()};
    }

  return genom_ok;
}


/* --- Function get_depth_stream_sizes ---------------------------------- */

/** Codel rs_get_stream_sizes of function get_depth_stream_sizes.
 *
 * Returns genom_ok.
 * Throws realsense_einval.
 */
/* already defined in service get_video_stream_sizes */



/* --- Function stream_video -------------------------------------------- */

/** Codel rs_stream_video of function stream_video.
 *
 * Returns genom_ok.
 * Throws realsense_einval, realsense_e_rs.
 */
genom_event
rs_stream_video(realsense_stream type, int16_t index, int16_t width,
                int16_t height, or_sensor_pixmap_encoding format,
                int16_t fps, realsense_config **config,
                realsense_pipeline **rspipe, const char name[16],
                const camera_pipeline *pipeline,
                const camera_worker_lock *mutex,
                const realsense_frame_queue *mqueue,
                realsense_streammap **streammap,
                const genom_context self)
{
  return rs_stream_enable(
    type, index, width, height, format, fps,
    config, rspipe, name, pipeline, mqueue, streammap, self);
}


/* --- Function stream_depth -------------------------------------------- */

/** Codel rs_stream_depth of function stream_depth.
 *
 * Returns genom_ok.
 * Throws realsense_einval, realsense_e_rs.
 */
genom_event
rs_stream_depth(realsense_stream type, int16_t index, int16_t width,
                int16_t height, or_sensor_pixmap_encoding format,
                int16_t fps, realsense_config **config,
                realsense_pipeline **rspipe, const char name[16],
                const camera_pipeline *pipeline,
                const camera_worker_lock *mutex,
                const realsense_frame_queue *mqueue,
                realsense_streammap **streammap,
                const genom_context self)
{
  switch(format) {
    case or_sensor_pixmap_ANY: format = or_sensor_pixmap_Z32; break;
    case or_sensor_pixmap_Z32:
    case or_sensor_pixmap_XYZ32: break;
    default: return realsense_einval(self);
  }

  return rs_stream_enable(
    type, index, width, height, format, fps,
    config, rspipe, name, pipeline, mqueue, streammap, self);
}


/* --- Function stream_motion ------------------------------------------- */

/** Codel rs_stream_motion of function stream_motion.
 *
 * Returns genom_ok.
 * Throws realsense_einval, realsense_e_rs.
 */
genom_event
rs_stream_motion(realsense_stream type, int16_t index, int16_t fps,
                 realsense_config **config,
                 realsense_pipeline **rspipe,
                 const camera_pipeline *pipeline,
                 const camera_worker_lock *mutex,
                 const realsense_frame_queue *mqueue,
                 realsense_streammap **streammap,
                 const genom_context self)
{
  return rs_stream_enable(
    type, index, 0, 0, or_sensor_pixmap_ANY, fps,
    config, rspipe, "", pipeline, mqueue, streammap, self);
}


/* --- rs_stream_enable ---------------------------------------------------- */

static genom_event
rs_stream_enable(realsense_stream type, int16_t index,
                int16_t width, int16_t height,
                or_sensor_pixmap_encoding format, int16_t fps,
                realsense_config **config,
                realsense_pipeline **rspipe,
                const char name[16], const camera_pipeline *pipeline,
                const realsense_frame_queue *mqueue,
                realsense_streammap **streammap,
                const genom_context self)
{
  rs2::stream_profile p, p0;
  bool stopped = true;

  try {
    /* stop devices */
    try { (*rspipe)->stop(); }
    catch(const rs2::wrong_api_call_sequence_error &) {
      /* OK: not started yet */
      stopped = false;
    }

    switch(type) {
      case realsense_STREAM_COLOR:
      case realsense_STREAM_INFRARED:
        (*config)->enable_stream(rscvt_stream(type), index, width, height,
                                 rscvt_format(format), fps);
        break;

      case realsense_STREAM_DEPTH:
        (*config)->enable_stream(rscvt_stream(type), index, width, height,
                                 RS2_FORMAT_ANY, fps);
        break;

      case realsense_STREAM_GYRO:
      case realsense_STREAM_ACCEL:
        (*config)->enable_stream(rscvt_stream(type), index, 0, 0,
                                 RS2_FORMAT_ANY, fps);
        break;

      default: /*NOTREACHED*/ return realsense_einval(self);
    }

    auto &&profile = (*config)->resolve(**rspipe);

    /* update common color profile for frame alignment */
    try { p0 = profile.get_stream(RS2_STREAM_COLOR); }
    catch (const std::runtime_error &e) { /* ignore if not present */ }
    for (auto &i : **streammap) i.second.p0 = p0;

    /* look for the new stream_profile and record a camera_job for it */
    p = profile.get_stream(rscvt_stream(type), index);
    switch(type) {
      case realsense_STREAM_COLOR:
      case realsense_STREAM_INFRARED:
      case realsense_STREAM_DEPTH:
        /* create job */
        realsense_job &job = (**streammap)[p];
        job.reset();
        job.format =
          format == or_sensor_pixmap_ANY ? rscvt_format(p.format()) : format;
        job.p0 = p0;
        job.sink(name);
        break;
    }
    warnx("enabled %s stream", p.stream_name().c_str());
  }

  catch(const rs2::error &e) {
    realsense_e_rs_detail d;
    snprintf(d.what, sizeof(d.what), "%s", e.what());

    /* re-enable previous streams */
    (*config)->disable_all_streams();
    for(const auto &kv : **streammap)
      (*config)->enable_stream(kv.first.type, kv.first.index);

    if (stopped)
      rs_stream_start(config, rspipe, pipeline, mqueue, streammap, self);

    return realsense_e_rs(&d,self);
  }

  /* restart devices */
  return rs_stream_start(config, rspipe, pipeline, mqueue, streammap, self);
}


/* --- rs_stream_start ----------------------------------------------------- */

static genom_event
rs_stream_start(realsense_config **config, realsense_pipeline **rspipe,
                const camera_pipeline *pipeline,
                const realsense_frame_queue *mqueue,
                realsense_streammap **streammap, const genom_context self)
{
  try {
    (*rspipe)->start(
      **config,
      [pipeline, streammap, mqueue](rs2::frame f) {
        if (f.is<rs2::motion_frame>())
          mqueue->enqueue(f);
        else {
          for(auto f : f.as<rs2::frameset>()) {
            const auto &v = f.as<rs2::video_frame>();
            if (!v) continue;

            const auto &p = v.get_profile().as<rs2::video_stream_profile>();
            realsense_job &job = (**streammap)[p];
            job.frames.enqueue(v);
            job.schedule(pipeline);
          }
        }
      });
  }

  catch(const rs2::error &e) {
    realsense_e_rs_detail d;
    snprintf(d.what, sizeof(d.what), "%s", e.what());
    return realsense_e_rs(&d,self);
  }

  return genom_ok;
}
