/*
 * Copyright (c) 2019,2023-2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                             Martin Jacquet - September 2022
 */

#include "acrealsense.h"

#include <err.h>

#include <algorithm>

#include "realsense_c_types.h"

#include "codels.hh"


/* --- Task main -------------------------------------------------------- */


/** Codel rs_start of task main.
 *
 * Triggered by realsense_start.
 * Yields to realsense_poll.
 */
genom_event
rs_start(realsense_ids *ids, const realsense_accel *accel,
         const realsense_gyro *gyro, const genom_context self)
{
  rs2::context ctx;

  ids->rspipe = new realsense_pipeline;
  ids->config = new realsense_config;
  ids->streammap = new realsense_streammap;
  ids->mqueue = new realsense_frame_queue;
  ids->mframe = new realsense_frame;

  for(int i = 0; i < 3; i++)
    ids->gstddev[i] = ids->astddev[i] = 0.1;

  /* reset any connected device */
  try {
    for(auto d : ctx.query_devices()) d.hardware_reset();
  } catch(const rs2::error &e) { warnx("%s", e.what()); }

  *accel->data(self) = or_pose_estimator_state{0};
  *gyro->data(self) = or_pose_estimator_state{0};

  return realsense_poll;
}


/** Codel rs_poll of task main.
 *
 * Triggered by realsense_poll.
 * Yields to realsense_poll, realsense_publish.
 */
genom_event
rs_poll(const realsense_frame_queue *mqueue, realsense_frame **mframe,
        const genom_context self)
{
  try {
    if (mqueue->try_wait_for_frame<rs2::frame>(*mframe, 500))
      return realsense_publish;
  }
  catch(const rs2::error &e) { warnx("%s", e.what()); }

  /* timeout or error */
  return realsense_poll;
}


/** Codel rs_publish of task main.
 *
 * Triggered by realsense_publish.
 * Yields to realsense_poll.
 */
genom_event
rs_publish(const realsense_frame *mframe, const double gstddev[3],
           const double astddev[3], const realsense_accel *accel,
           const realsense_gyro *gyro, const genom_context self)
{
  const rs2::motion_frame &m = mframe->as<rs2::motion_frame>();
  or_pose_estimator_state *port;
  double ts;

  if (!m) return realsense_poll;
  const auto &p = m.get_profile().as<rs2::motion_stream_profile>();

  switch(p.stream_type()) {
    case RS2_STREAM_ACCEL: port = accel->data(self); break;
    case RS2_STREAM_GYRO: port = gyro->data(self); break;
    default: /*NOTREACHED*/ return realsense_poll;
  }

  ts = m.get_timestamp() / 1000.;
  port->ts.sec = ts;
  port->ts.nsec = (ts - port->ts.sec) * 1000000000;
  port->intrinsic = true;

  const rs2_vector &imu = m.get_motion_data();
  switch(p.stream_type()) {
    case RS2_STREAM_ACCEL:
      port->acc._present = true;
      port->acc._value.ax = imu.x;
      port->acc._value.ay = imu.y;
      port->acc._value.az = imu.z;

      /* the IMU error model from rs2::video_stream_profile::get_intrinsics
       * seems to always be 0, so use a user-defined value */
      port->acc_cov._present = true;
      port->acc_cov._value.cov[0] = astddev[0] * astddev[0];
      port->acc_cov._value.cov[2] = astddev[1] * astddev[1];
      port->acc_cov._value.cov[5] = astddev[2] * astddev[2];

      accel->write(self);
      break;

    case RS2_STREAM_GYRO:
      port->avel._present = true;
      port->avel._value.wx = imu.x;
      port->avel._value.wy = imu.y;
      port->avel._value.wz = imu.z;

      /* the IMU error model from rs2::video_stream_profile::get_intrinsics
       * seems to always be 0, so use a user-defined value */
      port->avel_cov._present = true;
      port->avel_cov._value.cov[0] = gstddev[0] * gstddev[0];
      port->avel_cov._value.cov[2] = gstddev[1] * gstddev[1];
      port->avel_cov._value.cov[5] = gstddev[2] * gstddev[2];

      gyro->write(self);
      break;

    default: /*NOTREACHED*/ return realsense_poll;
  }

  return realsense_poll;
}


/** Codel rs_stop of task main.
 *
 * Triggered by realsense_stop.
 * Yields to realsense_ether.
 */
genom_event
rs_stop(realsense_ids *ids, const genom_context self)
{
  /* stop rs pipeline */
  try { ids->rspipe->stop(); }
  catch(const rs2::wrong_api_call_sequence_error &) {
    /* OK: was not started yet */
  }

  delete ids->rspipe;
  delete ids->mframe;
  delete ids->mqueue;
  delete ids->config;
  delete ids->streammap;

  return realsense_ether;
}


/* --- realsense_job::operator() ------------------------------------------- */

/* convert rs2::frame into a or_image_frame. This actually runs in a
 * camera worker thread context. */
camera_job::state
realsense_job::operator()()
{
  or_image_frame *image = out(0) ? out(0)->image : nullptr;
  rs2::frame f;

  if (!image) return idle;
  if (!frames.poll_for_frame(&f)) return idle;

  const auto &p = f.get_profile().as<rs2::video_stream_profile>();
  const auto &i = p.get_intrinsics();

  double ts = f.get_timestamp() / 1000.;
  image->ts.sec = ts;
  image->ts.nsec = (ts - image->ts.sec) * 1000000000;

  /* frame info */
  if ((uint32_t)p.unique_id() != image->info.seq) {
    try {
      image->info.seq = p.unique_id();
      image->info.roi._present = false;

      image->info.distortion._present = false;	/* XXX TBD */

      image->info.intrinsics._present = true;
      image->info.intrinsics._value.fx = i.fx;
      image->info.intrinsics._value.fy = i.fy;
      image->info.intrinsics._value.cx = i.ppx;
      image->info.intrinsics._value.cy = i.ppy;

      try {
        const auto &e = p.get_extrinsics_to(p0);
        const float * const R = e.rotation;		/* column major */
        const double R0 = e.rotation[0] / 4.;
        const double R4 = e.rotation[4] / 4.;
        const double R8 = e.rotation[8] / 4.;

        image->info.extrinsics._present = true;
        image->info.extrinsics._value.T.x = e.translation[0];
        image->info.extrinsics._value.T.y = e.translation[1];
        image->info.extrinsics._value.T.z = e.translation[2];
        image->info.extrinsics._value.R.qw =
          sqrt(std::max({0., 0.25 + R0 + R4 + R8}));
        image->info.extrinsics._value.R.qx =
          copysign(sqrt(std::max({0., 0.25 + R0 - R4 - R8})), R[5] - R[7]);
        image->info.extrinsics._value.R.qy =
          copysign(sqrt(std::max({0., 0.25 - R0 + R4 - R8})), R[6] - R[2]);
        image->info.extrinsics._value.R.qz =
          copysign(sqrt(std::max({0., 0.25 - R0 - R4 + R8})), R[1] - R[3]);
      }
      catch(rs2::error &) {
        image->info.extrinsics._present = false;
      }
    }
    catch(rs2::not_implemented_error &) {}
  }

  /* frame pixmap */
  image->pixmap.seq = f.get_frame_number();
  image->pixmap.width = i.width;
  image->pixmap.height = i.height;
  switch(format) {
    case or_sensor_pixmap_Z32:
      image->pixmap.stride = image->pixmap.width * sizeof(float); break;
    case or_sensor_pixmap_XYZ32:
      image->pixmap.stride = 3 * image->pixmap.width * sizeof(float); break;
    default:
      image->pixmap.stride = f.as<rs2::video_frame>().get_stride_in_bytes();
  }
  image->pixmap.pixfmt = format;

  uint32_t len = image->pixmap.stride * image->pixmap.height;
  if (image->pixmap.data._maximum < len) {
    if (genom_sequence_reserve(&image->pixmap.data, len) == -1)
      throw std::bad_alloc();
  }
  image->pixmap.data._length = len;

  union { uint8_t *u; float *f; } px;
  px.u = image->pixmap.data._buffer;
  switch(format) {
    case or_sensor_pixmap_Z32: {
      const auto &depth = f.as<rs2::depth_frame>();
      for(uint32_t v = 0; v < image->pixmap.height; v++)
        for(uint32_t u = 0; u < image->pixmap.width; u++) {
          *(px.f++) = depth.get_distance(u, v);
        }
      break;
    }

    case or_sensor_pixmap_XYZ32: {
      const auto &depth = f.as<rs2::depth_frame>();
      double z;
      for(uint32_t v = 0; v < image->pixmap.height; v++)
        for(uint32_t u = 0; u < image->pixmap.width; u++) {
          z = depth.get_distance(u, v);
          *(px.f++) = z * (u - i.ppx) / i.fx;
          *(px.f++) = z * (v - i.ppy) / i.fy;
          *(px.f++) = z;
        }
      break;
    }

    default:
      memcpy(px.u, f.get_data(), len);
      break;
  }

  return update;
}
